/*	$Id: action.c,v 1.2 1999/12/08 03:43:56 leonard Exp $	*/
/* Copyright 1999, David Leonard. All rights reserved. */

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>

#include <sys/types.h>
#include <sys/time.h>

#include "vox.h"
#include "action.h"
#include "run.h"

/*
 * Read some characters from the modem.
 * Returns NULL if some characters were received, or 
 * a command pointer if it timed out or an on condition was
 * matched.
 */
struct command *
readmodem(cp, lenp, nomatch)
	char *cp;
	size_t *lenp;
	int nomatch;
{
	fd_set	r;
	int	ret;
	struct command *res;
	int i;

	FD_ZERO(&r);
    again:
	FD_SET(modemfd, &r);
	ret = select(modemfd + 1, &r, NULL, NULL, 
	    timeout_target ? &timeout_remaining : NULL);

	if (ret == -1) {
	    if (errno == EAGAIN)
		goto again;
	    err(1, "select");
	}

	if (ret == 0) {
		if (verbose)
			fprintf(stderr, "timer expired\n");
		res = timeout_target;
		timeout_target = NULL;	/* clear timer */
		return res;
	}

	if (!FD_ISSET(modemfd, &r))
		errx(1, "readmodem: internal error");

	ret = read(modemfd, cp, *lenp);
	if (ret == -1)
		err(1, "read");
	if (ret == 0)
		errx(1, "read: closed");
	*lenp = ret;
	if (nomatch == 0) {
		if (verbose) {
			fprintf(stderr, "read ");
			print_escaped(stderr, cp, ret);
			fprintf(stderr, "\n");
		}
		for (i = 0; i < ret; i++) {
			res = checkmatch(cp[i]);
			if (res)
				return res;
		}
	}
	return NULL;
}

/*
 * Write some data to the modem, timing out
 * if it doesn't become ready, and reading data
 * as we go. Returns NULL if successfully written,
 * or a command pointer if a timeout occurred or 
 * some data was received that matched an 'on' condition.
 */
struct command *
writemodem(data, lenp)
	char *data;
	size_t *lenp;
{
	fd_set	r, w;
	int	ret;
	int	left;
	struct command *res;
	char	c;
	
	FD_ZERO(&w);
	FD_ZERO(&r);
	left = *lenp;
	res = NULL;
	while (left) {
		FD_SET(modemfd, &w);
		FD_SET(modemfd, &r);
		ret = select(modemfd + 1, &r, &w, NULL, 
		    timeout_target ?  &timeout_remaining : NULL);

		if (ret == -1) {
		    if (errno == EAGAIN)
			continue;
		    err(1, "select");
		}

		if (ret == 0) {
			if (verbose)
				fprintf(stderr, "timer expired\n");
			res = timeout_target;
			timeout_target = NULL;	/* clear timer */
			break;
		}

		if (FD_ISSET(modemfd, &w)) {
			ret = write(modemfd, data, left);
			if (ret == -1)
				err(1, "write");
			if (ret == 0)
				errx(1, "write: zero");
			data += ret;
			left -= ret;
		}

		if (FD_ISSET(modemfd, &r)) {
			ret = read(modemfd, &c, sizeof c);
			if (ret == -1)
				err(1, "read");
			if (ret == 0)
				errx(1, "read: closed");
			res = checkmatch(c);
		}
	}
	*lenp -= left;
	return res;
}

/*
 * Set up the timeout timer
 */
void
start_timer(delay)
	int delay;
{
	timeout_remaining.tv_sec = delay;
	timeout_remaining.tv_usec = 0;

	if (delay > 0 && timeout_target == NULL)
		warnx("timer should be started after an 'on timeout'");
	if (delay == 0)
		timeout_target = NULL;
}

/*
 * Read and discard data until either the 
 * string we are expecting arrives, or an on condition is matched
 */
struct command *
waitstr(str, next)
	char * str;
	struct command * next;
{
	struct command this;
	struct command * res;
	int	added = 0;
	char	c;
	size_t	sz;

	if (str && *str == 0)
		return next;

	if (str) {
		added = addmatch(str, &this);
		if (verbose) {
			fprintf(stderr, "waiting for ");
			print_escaped(stderr, str, strlen(str));
			fprintf(stderr, "\n");
		}
	}
	do {
		sz = sizeof c;
		res = readmodem(&c, &sz, 0);
	} while (res == NULL);
	if (str && added)
		removematch();
	if (res == &this)
		res = next;
	return res;
}

/*
 * Consume characters until an on-condition arrives, or the
 * delay given passes.
 */
struct command *
delay(delay, next)
	int delay;
	struct command * next;
{
	char c;
	size_t sz;
	struct command * ret;
	struct timeval delay_tv;
	struct timeval saved_remaining;
	struct command * saved_target;

	if (delay == 0)
		return next;

	delay_tv.tv_sec = delay;
	delay_tv.tv_usec = 0;

	/*
	 * If our delay is longer than the pending timeout, 
	 * just spin, reading data. 
	 */
	if (timeout_target && timercmp(&timeout_remaining, &delay_tv, <= )) {
		do {
			sz = sizeof c;
			ret = readmodem(&c, &sz, 0);
		} while (ret == NULL);
		return ret;
	}

	/* Save the old timeout */
	saved_target = timeout_target;
	memcpy(&saved_remaining, &timeout_remaining, sizeof saved_remaining);

	/* Set up the new timeout */
	timeout_target = next;
	memcpy(&timeout_remaining, &delay_tv, sizeof timeout_remaining);

	/* Consume */
	do {
		sz = sizeof c;
		ret = readmodem(&c, &sz, 0);
	} while (ret == NULL);

	/* Restore the old timeout */
	timeout_target = saved_target;
	if (timeout_target) {
		/*
		 * Compute how much time we spent in readmodem()
		 * and remove it from the timer; ie timout_remaining = 
		 * (saved_remaining - (delay_tv - timeout_remaining)) 
		 */
		timersub(&delay_tv, &timeout_remaining, &delay_tv);
		timersub(&saved_remaining, &delay_tv, &timeout_remaining);
	}

	return ret;
}

struct command *
quote(text, next)
	char *text;
	struct command *next;
{
	struct command *res;
	size_t len;

	len = strlen(text);
	if (verbose) {
		fprintf(stderr, "sending ");
		print_escaped(stderr, text, len);
		fprintf(stderr, "\n");
	}
	res = writemodem(text, &len);
	if (res == NULL)
		return next;
	else
		return res;
}

struct command *
vcon(next)
	struct command *next;
{
	return waitstr("VCON\r\n", next);
}
