/*	$Id: audio.c,v 1.3 1999/12/08 03:45:29 leonard Exp $	*/
/* Copyright 1999, David Leonard. All rights reserved. */

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <err.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include "vox.h"
#include "action.h"
#include "run.h"

/* RMD (Raw Modem Data) header format from mgetty */
typedef struct {
	char	magic[4];
	char	voice_modem_type[16];
	short	compression;
	short	speed;
	char	bits;
	char	reserved[7];
} rmd_header;

#define BURST 8192

char dledle[] = { DLE, DLE };
char dlecan[] = { DLE, CAN };
char dleetx[] = { DLE, ETX };

/*
 * Play the audio from a file to the modem.
 * DLE escaping is done in this function after reading from the file.
 * Assumes the modem has already been sent AT#VTX command.
 * Sends DLE-ETX at end of successful play.
 * If an on-condition arises, sends DLE-CAN before branching.
 */
struct command *
play(filename, next)
	char * filename;
	struct command * next;
{
	struct command *res;
	struct stat st;
	char *base = MAP_FAILED;
	off_t len;
	char *pos;
	int fd;
	size_t wlen;
	rmd_header *hdr;

	if ((fd = open(filename, O_RDONLY)) < 0)
		goto fail;
	if (fstat(fd, &st) < 0)
		goto fail;
	base = mmap(NULL, st.st_size, PROT_READ, MAP_FILE, fd, 0);
	if (base == MAP_FAILED)
		goto fail;
	len = st.st_size;

	if (len < sizeof (rmd_header)) {
		warn("%s: too short", filename);
		res = next;
		goto out;
	}

	/* Skip RMD header */
	hdr = (rmd_header *)base;
	pos = base + sizeof (rmd_header);
	len -= sizeof(rmd_header);

	if (memcmp(hdr->magic, "RMD1", sizeof hdr->magic) != 0)  {
		warnx("%s: not an rmd file", filename);
		/* res = next; goto out; */
	}

	/* Send VTX command, wait for CONNECT */
	res = quote("AT#VTX\r", NULL);
	if (res != NULL)
		goto out;

	res = waitstr("CONNECT\r\n", NULL);

	if (verbose)
		fprintf(stderr, "playing file %s (%d bytes)\n", filename,
		    (int)len);

	while (len) {
		if (res != NULL)  {
			if (verbose)
				fprintf(stderr, "sending DLE CAN DLE ETX\n");
			write(modemfd, dlecan, sizeof dlecan);
			write(modemfd, dleetx, sizeof dleetx);
			res = vcon(res);
			goto out;
		}

		if (*pos == DLE) {
			wlen = sizeof dledle;
			res = writemodem(dledle, &wlen);
			wlen = 1;
		} else {
			/* Write either to end of data or to next DLE */
			for (wlen = 1; wlen < len && wlen < BURST; wlen++)
				if (pos[wlen] == DLE)
					break;
			res = writemodem(pos, &wlen);
		}

		len -= wlen;
		pos += wlen;
	}

	if (verbose)
		fprintf(stderr, "sending DLE ETX\n");
	write(modemfd, dleetx, sizeof dleetx);		/* end stream */
	res = vcon(next);
	goto out;

    fail:
	warn("play: %s", filename);
	res = next;

    out:
	if (fd != -1) {
		if (base != MAP_FAILED)
			munmap(base, st.st_size);
		close(fd);
	}
	return res;
}

/*
 * Record the audio the modem into a file.
 * DLE escaping is done in this function before saving to the file.
 * Assumes the modem has been sent the #VRX command.
 * Stops when DLE-ETX is received, on condition or an i/o error.
 * If an on-condition arises, sends DLE-CAN before branching.
 */
struct command *
record(filename, next)
	char * filename;
	struct command * next;
{
	char buffer[8192];
	char *pos, *spos;
	int fd;
	int len;
	int indle;
	struct command * res;
	rmd_header hdr;

	/* Send VRX command, wait for CONNECT */
	res = quote("AT#VRX\r", NULL);
	if (res != NULL)
		return res;
	res = waitstr("CONNECT\r\n", NULL);
	if (res != NULL)
		return res;

	fd = open(filename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
	if (fd < 0)  {
		warn("record: %s", filename);
		return next;
	}

	/* Write the RMD header for a rockwell */
	memcpy(hdr.magic, "RMD1", sizeof hdr.magic);
	strlcpy(hdr.voice_modem_type, "Rockwell", sizeof hdr.voice_modem_type);
	hdr.compression = htons(4);
	hdr.speed = htons(7200);
	hdr.bits = 4;
	memset(hdr.reserved, '\0', sizeof hdr.reserved);
	if (write(fd, &hdr, sizeof hdr) == -1) {
		warn("write: %s", filename);
		res = next;
	}

	/* Start decoding the modem data */
	indle = 0;
	while (res == NULL) {
		/* Read some stuff */
		len = sizeof buffer;
		res = readmodem(buffer, &len, 1);

		/* Unstuff DLEs */
		pos = buffer;
		while (len) {
			spos = pos;
			if (!indle && *pos == DLE) {
				/* Found a DLE character: save & reloop. */
				pos++;
				len--;
				indle = 1;
				continue;
			}
			if (indle) {
				/* We are expecting the char after a dle */
				indle = 0;
				if (*pos == SUB) {
					if (write(fd, dledle, sizeof dledle)
					    == -1) {
						warn("record: %s", filename);
						res = next;
						break;
					}
					pos++;
					len--;
					continue;
				} else if (*pos == DLE) {
					pos++;	
					len--;
					/* fall through to consume DLE */
				} else if (*pos == ETX) {
					/* Successful receive */
					res = next;
					break;
				} else {
					if (verbose) {
						if (*pos < ' ' || *pos >= 0x7f)
							fprintf(stderr, 
							    "got DLE \\%03o\n",
							    *pos);
						else
							fprintf(stderr, 
							    "got DLE '%c'\n",
							    *pos);
					}
					/* Pass these to the matcher */
					checkmatch(DLE); /* XXX */
					res = checkmatch(*pos);
					if (res)
						break;
					pos++;
					len--;
					continue;
				}
			} 
			while (len && *pos != DLE) {
				pos++;
				len--;
			}
			if (write(fd, spos, pos - spos) == -1) {
				warn("record: %s", filename);
				res = next;
				break;
			}
		}
	} 

	/*
	 * To stop record, simply send a 'key abort', which is
	 * tech speak for any character.
	 */
	if (verbose)
		fprintf(stderr, "sending \"!\"\n");
	write(modemfd, "!", 1);
	close(fd);

	/* XXX could also receive an OK response here? */
	return vcon(res);
}
