/*	$Id: run.c,v 1.2 1999/12/08 03:49:00 leonard Exp $	*/
/* Copyright 1999, David Leonard. All rights reserved. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/time.h>

#include "vox.h"
#include "run.h"
#include "action.h"
#include "exec.h"
#include "audio.h"

struct match {
	char *	str;
	int	len;
	struct command *target;
};

static void clearmatch(void);

static char *	match_history = NULL;
static int	match_history_len = 0;
static int 	nmatches = 0;
static int 	maxmatches = 0;
static struct match * matches = NULL;

struct command *timeout_target;
struct timeval timeout_remaining;
int 	verbose = 0;
volatile int exitting = 0;
volatile struct command *exit_target = NULL;

/*
 * Add an on-condition to the list of matches we look for
 */
int
addmatch(str, target)
	char *str;
	struct command *target;
{
	struct match *m;
	char *	newhist;
	int i;

	/* Check to see if the string/target is already in the table */
	for (i = 0; i < nmatches; i++)
		if (matches[i].target == target &&
		    strcmp(matches[i].str, str) == 0)
			return 0;

	nmatches++;
	if (nmatches > maxmatches) {
		if (maxmatches == 0) {
			maxmatches = 64;
			matches = malloc(maxmatches * sizeof *matches);
		} else {
			maxmatches *= 2;
			matches = realloc(matches, 
			    maxmatches * sizeof *matches);
		}
		if (matches == NULL)
			err(1, "malloc");
	}
	m = &matches[nmatches - 1];
	m->str = str;
	m->len = strlen(str);
	m->target = target;

	if (match_history_len < m->len) {
		newhist = malloc(m->len);
		if (newhist == NULL)
			err(1, "malloc");
		memset(newhist, '\0', m->len - match_history_len);
		if (match_history)
			memcpy(newhist + m->len - match_history_len,
			    match_history, match_history_len);
		match_history = newhist;
		match_history_len = m->len;
	}

	return 1;
}

/*
 * Remove the last added match - used for the 'wait' command
 */
void
removematch()
{
	nmatches--;
	if (nmatches == 0)
		clearmatch();
}

/*
 * Record a character in the match history and return a 
 * command target if any on-condition now matches it.
 */
struct command *
checkmatch(c)
	char c;
{
	struct match *m;
	int i;

	if (nmatches == 0)
		return NULL;

	/* Add character to history */
	memmove(match_history, match_history + 1, match_history_len - 1);
	match_history[match_history_len - 1] = c;

	/* Match last on-condition first */
	for (i = nmatches - 1; i >= 0; i--) {
		m = &matches[i];
		if (memcmp(&match_history[match_history_len - m->len],
		    m->str, m->len) == 0) {
			if (verbose) {
				fprintf(stderr, "matched ");
				if (m->str[0] == DLE) {
					fprintf(stderr, "dle ");
					print_escaped(stderr, m->str + 1, 
					    m->len - 1);
				} else
					print_escaped(stderr, m->str, m->len);
				fprintf(stderr, "\n");
			}
			return m->target;
		}
	}
	return NULL;
}

/*
 * Clear the table of on-conditions and the match history
 */
static void
clearmatch()
{

	nmatches = 0;
	if (match_history) {
		free(match_history);
		match_history = NULL;
		match_history_len = 0;
	}
	timeout_target = NULL;
}

/* 
 * Run a script.
 */
void
run(start)
	struct command *start;
{
	struct command *cmd;
	struct command *next;

	cmd = start;
	while (cmd) {
		next = cmd->next;

		if (verbose)
			print_command(stdout, cmd);

		switch (cmd->action) {
		case CMD_ON_CLEAR:
			clearmatch();
			timeout_target = NULL;
			break;
		case CMD_ON_STR:
			(void)addmatch(cmd->string, cmd->dest.cmd);
			break;
		case CMD_ON_TIMEOUT:
			timeout_target = cmd->dest.cmd;
			break;
		case CMD_LOCK:
			lock(cmd->onoff);
			break;
		case CMD_TIMER:
			if (cmd->number)
				start_timer(cmd->number);
			else
				timeout_target = NULL;
			break;
		case CMD_QUOTE:
			next = quote(cmd->string, next);
			break;
		case CMD_WAIT:
			next = waitstr(cmd->string, next);
			break;
		case CMD_VCON:
			next = vcon(next);
			break;
		case CMD_DELAY:
			next = delay(cmd->number, next);
			break;
		case CMD_GOTO:
			next = cmd->dest.cmd;
			break;
		case CMD_PLAY:
			next = play(cmd->string, next);
			break;
		case CMD_RECORD:
			next = record(cmd->string, next);
			break;
		case CMD_EXEC:
			do_exec(cmd->string);
			break;
		case CMD_SYSLOG:
			syslog(LOG_INFO, "%s", cmd->string);
			break;
		case CMD_SETENV:
			setenv(cmd->envvar, cmd->string, 1);
			break;
		default:
			errx(1, "unknown action %d", cmd->action);
		}
		cmd = next;

		if (exitting) {
			cmd = (struct command *)exit_target;
			exitting = 0;
		}
	}
}

